# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function

from datetime import datetime
import os
import sys

from accelerator.colourwrapper import colour
from accelerator.job import Job
from accelerator.metadata import extract_metadata, b64hash_setup
from .parser import ArgumentParser


def validate(data):
	warnings = []
	job = data.job.rsplit('/', 1)[-1]
	try:
		job = Job(job)
		job.path # may trigger NoSuchWorkdirError
	except Exception:
		job = None
	if job:
		if job.path != data.job:
			warnings.append("path mismatch (%r != %r)" % (job.path, data.job,))
	else:
		warnings.append("unknown job")
	h = b64hash_setup(data.job + '/setup.json')
	if h != data.setup_hash:
		if h:
			warnings.append("does not match job on disk")
		else:
			warnings.append("does not exist on disk")
	return warnings

def main(argv, cfg):
	descr = "show which job produced a file, for files downloaded through board."
	parser = ArgumentParser(
		prog=argv.pop(0),
		description=descr,
	)
	parser.add_argument('-v', '--verbose', action='store_true', negation='not', help='more output')
	parser.add_argument('filename', nargs='+')
	args = parser.parse_intermixed_args(argv)
	res = 0
	os.chdir(cfg.user_cwd)
	for filename in args.filename:
		if len(args.filename) > 1:
			prefix = filename + ': '
		else:
			prefix = ''
		with open(filename, 'rb') as fh:
			found = extract_metadata(filename, fh)
			if not found:
				res = 1
				print(prefix + colour('UNSUPPORTED FORMAT', 'sherlock/warning'), file=sys.stderr)
				continue
			try:
				found = list(found)
			except Exception:
				found = [None]
			if not found:
				res = 1
				if args.verbose or len(args.filename) > 1:
					print(prefix + colour('no data found', 'sherlock/warning'), file=sys.stderr)
				continue
			for data in found:
				if data is None:
					res = 1
					print(prefix + colour('decoding error', 'sherlock/warning'), file=sys.stderr)
					continue
				if args.verbose:
					print('%s%s' % (prefix, data.job,), end='')
				else:
					print('%s%s' % (prefix, data.job.rsplit('/', 1)[-1],), end='')
				warnings = validate(data)
				if warnings:
					print(' ' + colour(', '.join(warnings), 'sherlock/warning'), end='')
				if args.verbose:
					ts = datetime.fromtimestamp(data.time)
					print(' (%s at %s on %s)' % (data.method, ts, data.host), end='')
				print()
	return res



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function

from accelerator.error import AcceleratorError
from accelerator.job import Job
from accelerator.shell.parser import ArgumentParser

import os
import shutil
import sys

def main(argv, cfg):
	parser = ArgumentParser(prog=argv.pop(0), description='''delete stale and unused jobs''')
	parser.add_argument('-n', '--dry-run', action='store_true', negation='not', help="don't delete anything, just print")
	args = parser.parse_intermixed_args(argv)

	to_delete = []
	def candidate(jid):
		job = Job(jid)
		try:
			job.post
		except (OSError, AcceleratorError):
			to_delete.append(job)

	for name, path in cfg.workdirs.items():
		for jid in os.listdir(path):
			if '-' in jid:
				wd, num = jid.rsplit('-', 1)
				if wd == name and num.isdigit():
					candidate(jid)
	if not to_delete:
		print("Nothing to do.")
		return 0
	if args.dry_run:
		print("Would have deleted %d jobs" % (len(to_delete),))
		return 0
	print("Deleting %d jobs" % (len(to_delete),))
	for job in to_delete:
		shutil.rmtree(job.path)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Concatenate a dataset chain into a single dataset.

Reads datasets.source until datasets.previous: 'source', so you can chain
these jobs easily.

This is a performance optimisation, when you are willing to pay some time
now to avoid the cost of switching datasets later.
'''

datasets = ('source', 'previous',)

def prepare(job):
	chain = datasets.source.chain(stop_ds={datasets.previous: 'source'})
	columns = {name: col.type for name, col in datasets.source.columns.items()}
	hashlabel = datasets.source.hashlabel
	for ds in chain:
		if columns != {name: col.type for name, col in ds.columns.items()}:
			raise Exception('Dataset %s does not have the same columns as %s' % (ds.quoted, datasets.source.quoted,))
		if hashlabel != ds.hashlabel:
			raise Exception('Dataset %s has hashlabel %r, expected %r' % (ds.quoted, ds.hashlabel, hashlabel,))
	dw = job.datasetwriter(hashlabel=hashlabel, previous=datasets.previous, copy_mode=True)
	for name, t in sorted(columns.items()):
		dw.add(name, t, none_support=chain.none_support(name))
	return dw, chain

def analysis(sliceno, job, prepare_res):
	dw, chain = prepare_res
	write = dw.write
	for items in chain.iterate(sliceno, copy_mode=True):
		write(*items)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2023 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify various corner cases in dataset_type.
'''

from collections import defaultdict
from datetime import date, time, datetime, timedelta
from math import isnan
import json
import random
import struct
import sys

from accelerator import subjobs
from accelerator.dispatch import JobError
from accelerator.dataset import Dataset, DatasetWriter
from accelerator.dsutil import typed_writer
from accelerator.compat import PY3, UTC
from accelerator.standard_methods import dataset_type
from accelerator import g

options = {'numeric_comma': True}

depend_extra = (dataset_type,)

all_typenames = set(dataset_type.convfuncs)
used_typenames = set()

def used_type(typ):
	if ':' in typ and typ not in all_typenames:
		typ = typ.split(':', 1)[0] + ':*'
	used_typenames.add(typ)

no_default = object()
def verify(name, types, bytes_data, want, default=no_default, want_fail=False, all_source_types=False, exact_types=False, **kw):
	todo = [('bytes', bytes_data,)]
	if all_source_types:
		uni_data = [v.decode('ascii') for v in bytes_data]
		todo += [('ascii', uni_data,), ('unicode', uni_data,)]
	for coltype, data in todo:
		dsname = '%s %s' % (name, coltype,)
		_verify(dsname, types, data, coltype, want, default, want_fail, exact_types, kw)

def _verify(name, types, data, coltype, want, default, want_fail, exact_types, kw):
	if callable(want):
		assert not exact_types, "Don't provide a callable want with exact_types=True"
		check = want
	else:
		def check(got, fromstr, filtered=False):
			want1 = want if isinstance(want, list) else want[typ]
			if filtered:
				want1 = want1[::2]
			if exact_types:
				got = [(v, type(v).__name__) for v in got]
				want1 = [(v, type(v).__name__) for v in want1]
			assert got == want1, 'Expected %r, got %r from %s.' % (want1, got, fromstr,)
	dw = DatasetWriter(name=name, columns={'data': coltype, 'extra': 'bytes'})
	dw.set_slice(0)
	for ix, v in enumerate(data):
		dw.write(v, b'1' if ix % 2 == 0 else b'skip')
	for sliceno in range(1, g.slices):
		dw.set_slice(sliceno)
	bytes_ds = dw.finish()
	for typ in types:
		opts = dict(column2type=dict(data=typ))
		opts.update(kw)
		if default is not no_default:
			opts['defaults'] = {'data': default}
		try:
			jid = subjobs.build('dataset_type', datasets=dict(source=bytes_ds), options=opts)
		except JobError:
			if want_fail:
				continue
			raise Exception('Typing %r as %s failed.' % (bytes_ds, typ,))
		assert not want_fail, "Typing %r as %s should have failed, but didn't (%s)." % (bytes_ds, typ, jid)
		typed_ds = Dataset(jid)
		got = list(typed_ds.iterate(0, 'data'))
		check(got, '%s (typed as %s from %r)' % (typed_ds, typ, bytes_ds,))
		if opts.get('filter_bad'):
			bad_ds = Dataset(jid, 'bad')
			got_bad = list(bad_ds.iterate(0, 'data'))
			assert got_bad == [b'nah'], "%s should have had a single b'nah', but had %r" % (bad_ds, got_bad,)
		if 'filter_bad' not in opts and not callable(want):
			opts['filter_bad'] = True
			opts['column2type']['extra'] = 'int32_10'
			jid = subjobs.build('dataset_type', datasets=dict(source=bytes_ds), options=opts)
			typed_ds = Dataset(jid)
			got = list(typed_ds.iterate(0, 'data'))
			check(got, '%s (typed as %s from %r with every other line skipped from filter_bad)' % (typed_ds, typ, bytes_ds,), True)
			want_bad = [t for t in bytes_ds.iterate(0) if t[1] == b'skip']
			bad_ds = Dataset(jid, 'bad')
			got_bad = list(bad_ds.iterate(0))
			assert got_bad == want_bad, "Expected %r, got %r from %s" % (want_bad, got_bad, bad_ds,)
		used_type(typ)

def test_numbers():
	verify('floats', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'1.50', b'-inf', b'5e-1'], [1.5, float('-inf'), 0.5], all_source_types=True)
	if options.numeric_comma:
		verify('numeric_comma', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'1,5', b'1.0', b'9'], [1.5, 42.0, 9.0], '42', numeric_comma=True)
	verify('float32 rounds', ['complex32', 'float32'], [b'1.2'], [1.2000000476837158])
	verify('filter_bad', ['int32_10', 'int64_10', 'complex32', 'complex64', 'float32', 'float64', 'number'], [b'4', b'nah', b'1', b'0'], [4, 1, 0], filter_bad=True)

	all_source_types = True
	for base, values in (
		(10, (b'27', b'027', b' \r27 '),),
		( 8, (b'33', b'033', b'\t000033'),),
		(16, (b'1b', b'0x1b', b'\r001b',),),
		( 0, (b'27', b'\r033', b'0x1b',),),
	):
		types = ['%s_%d' % (typ, base,) for typ in ('int32', 'int64',)]
		verify('base %d' % (base,), types, values, [27, 27, 27], all_source_types=all_source_types)
		types = [typ + 'i' for typ in types]
		if base == 10:
			types += ['float32i', 'float64i']
		values = [v + b'garbage' for v in values]
		verify('base %d i' % (base,), types, values, [27, 27, 27], all_source_types=all_source_types)
		all_source_types = False
	# python2 has both int and long, let's not check exact types there.
	verify('inty numbers', ['number', 'number:int'], [b'42', b'42.0', b'42.0000000', b'43.', b'.0'], [42, 42, 42, 43, 0], exact_types=PY3)
	if options.numeric_comma:
		verify('inty numbers numeric_comma', ['number', 'number:int'], [b'42', b'42,0', b'42,0000000', b'43,', b',0'], [42, 42, 42, 43, 0], numeric_comma=True, exact_types=PY3)

	# Python 2 accepts 42L as an integer, python 3 doesn't. The number
	# type falls back to python parsing, verify this works properly.
	verify('integer with L', ['number'], [b'42L'], [42], want_fail=PY3)

	# tests both that values outside the range are rejected
	# and that None works as a default value.
	for typ, values, default in (
		('int32_10', (b'2147483648', b'-2147483648', b'1O',), '123',),
		('int32_16', (b'80000000', b'-80000000', b'1O',), None,),
		('int64_10', (b'36893488147419103231', b'9223372036854775808', b'-9223372036854775808', b'1O',), None,),
	):
		if default is None:
			want = [None] * len(values)
		else:
			want = [int(default)] * len(values)
		verify('nearly good numbers ' + typ, [typ], values, want, default, exact_types=True)

	verify('not a number', ['number'], [b'forty two'], [42], want_fail=True)
	verify('just a dot', ['number', 'float32'], [b'.'], [0], want_fail=True)

	verify('strbool', ['strbool'], [b'', b'0', b'FALSE', b'f', b'FaLSe', b'no', b'off', b'NIL', b'NULL', b'y', b'jao', b'well, sure', b' ', b'true'], [False] * 9 + [True] * 5, exact_types=True)
	verify('floatbool false', ['floatbool'], [b'0', b'-0', b'1', b'1004', b'0.00001', b'inf', b'-1', b' 0 ', b'0.00'], [False, False, True, True, True, True, True, False, False], exact_types=True)
	verify('floatbool i', ['floatbooli'], [b'1 yes', b'0 no', b'0.00 also no', b'inf yes', b' 0.01y', b''], [True, False, False, True, True, False], exact_types=True)
	def check_special(got, fromstr):
		msg = 'Expected [inf, -inf, nan, nan, nan, nan, inf], got %r from %s.' % (got, fromstr,)
		for ix, v in ((0, float('inf')), (1, float('-inf')), (-1, float('inf'))):
			assert got[ix] == v, msg
		for ix in range(2, 6):
			v = got[ix]
			if isinstance(v, complex):
				v = v.real
			assert isnan(v), msg
	verify('special floats', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'+Inf', b'-inF', b'nan', b'NaN', b'NAN', b'Nan', b'INF'], check_special)
	verify('complex', ['complex32', 'complex64'], [b'0.25-5j', b'-0.5j', b'infj'], [0.25-5j, -0.5j, complex(0, float('inf'))], exact_types=True)
	if options.numeric_comma:
		verify('complex numeric_comma', ['complex32', 'complex64'], [b'-0,5+1,5j', b'-0,5+1.5j', b',5j'], [-0.5+1.5j, 42j, 0.5j], '42j', numeric_comma=True, exact_types=True)

def test_bytes():
	want = [b'foo', b'\\bar', b'bl\xe4', b'\x00\t \x08', b'a\xa0b\x255\x00']
	for typs, data in (
		(['bytes', 'bytesstrip'], want,),
		(['bytesstrip'], [b'\t\t\t\rfoo\n \x0c', b'\\bar', b'bl\xe4   \x0b', b' \x00\t \x08', b'a\xa0b\x255\x00\n'],),
	):
		verify(typs[0], typs, data, want)

def test_ascii():
	for prefix, data in (
		('ascii', [b'foo', b'\\bar', b'bl\xe4', b'\x00\t \x08'],),
		('asciistrip', [b'\t\t\t\rfoo\n \x0c', b'\\bar', b'bl\xe4   \x0b', b' \x00\t \x08\t'],),
	):
		for encode, want in (
			(None,      ['foo', '\\bar',    'bl\\344', '\x00\t \x08'],),
			('replace', ['foo', '\\bar',    'bl\\344', '\x00\t \x08'],),
			('encode',  ['foo', '\\134bar', 'bl\\344', '\x00\t \x08'],),
			('strict',  ['foo', '\\bar',    None,      '\x00\t \x08'],),
		):
			if encode:
				typ = prefix + ':' + encode
			else:
				typ = prefix
			verify(typ, [typ], data, want, default=None)
	verify(
		'ascii all source types', ['ascii', 'asciistrip'],
		[b'\t\t\t\rfoo\r \x0c', b'\\bar', b' \x00\t \x08\t'],
		{
			'ascii': ['\t\t\t\rfoo\r \x0c', '\\bar', ' \x00\t \x08\t'],
			'asciistrip': ['foo', '\\bar', '\x00\t \x08'],
		},
		all_source_types=True,
	)

def test_unicode():
	data = [b'foo bar\n', b'\tbl\xe5', b'\tbl\xe5a \xe4  ', b'\tbl\xc3\xa5a ', b'\r+AOU-h']
	want = {
		'unicode:utf-8': ['foo bar\n', '既定 ', '既定 ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/strict': ['foo bar\n', '既定 ', '既定 ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tblåa ', '\r+AOU-h'],
		'unicode:iso-8859-1': ['foo bar\n', '\tblå', '\tblåa ä  ', '\tbl\xc3\xa5a ', '\r+AOU-h'],
		'unicode:ascii/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tbl\ufffd\ufffda ', '\r+AOU-h'],
		'unicode:ascii/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tbla ', '\r+AOU-h'],
		'unicode:utf-7/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tbla ', '\råh'],
		'unicode:utf-7/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tbl\ufffd\ufffda ', '\råh'],
		'unicodestrip:utf-8': ['foo bar', '既定', '既定', 'blåa', '+AOU-h'],
		'unicodestrip:utf-8/strict': ['foo bar', '既定', '既定', 'blåa', '+AOU-h'],
		'unicodestrip:utf-8/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'blåa', '+AOU-h'],
		# strip happens before ignore, so strip+ignore will have one space from '\tbl\xe5a \xe4  '
		'unicodestrip:utf-8/ignore': ['foo bar', 'bl', 'bla ', 'blåa', '+AOU-h'],
		'unicodestrip:iso-8859-1': ['foo bar', 'blå', 'blåa ä', 'bl\xc3\xa5a', '+AOU-h'],
		'unicodestrip:ascii/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'bl\ufffd\ufffda', '+AOU-h'],
		'unicodestrip:ascii/ignore': ['foo bar', 'bl', 'bla ', 'bla', '+AOU-h'],
		'unicodestrip:utf-7/ignore': ['foo bar', 'bl', 'bla ', 'bla', 'åh'],
		'unicodestrip:utf-7/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'bl\ufffd\ufffda', 'åh'],
	}
	verify('unicode', list(want), data, want, default='既定 ');
	# 既定 is not ascii, so we need a separate default for ascii-like types.
	want = {
		'unicode:ascii/strict': ['foo bar\n', 'standard', 'standard', 'standard', '\r+AOU-h'],
		'unicodestrip:ascii/strict': ['foo bar', 'standard', 'standard', 'standard', '+AOU-h'],
		'unicodestrip:utf-7/strict': ['foo bar', 'standard', 'standard', 'standard', 'åh'],
	}
	verify('unicode with ascii default', list(want), data, want, default='standard');
	verify('utf7 all', ['unicode:utf-7/replace'], [b'a+b', b'a+-b', b'+ALA-'], ['a\ufffd', 'a+b', '°'], all_source_types=True);

def test_datetimes():
	# These use the python datetime classes to verify valid dates,
	# so on Python < 3.6 they will accept various invalid dates.
	# So we only try nearly-good dates on Python 3.6+.

	# We start with some normal verify() calls which test basic function,
	# normal defaults and timezones.
	name = 'date YYYYMMDD'
	typ = 'date:%Y%m%d'
	data = [b'20190521', b'19700101', b'1970-01-01', b'1980', b'nah']
	want = [date(2019, 5, 21), date(1970, 1, 1), date(1945, 6, 20), date(1945, 6, 20), date(1945, 6, 20)]
	default = '19450620'
	verify(name, [typ], data, want, default, all_source_types=True)
	# A datei:-one with extra numbers at end (later i-versions will not add numbers)
	idata = [v + b'1868' for v in data]
	default += '42'
	verify(name + ' i', [typ.replace(':', 'i:', 1)], idata, want, default)

	# Timezone tests. I hope all systems accept the :Region/City syntax.
	verify('tz a', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 11:44', b'x'], [datetime(2020, 9, 30, 11, 44), datetime(2022, 11, 26, 23, 15)], default='2022-11-26 23:15')
	verify('tz b', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 11:44', b'x'], [datetime(2020, 9, 30, 11, 44, tzinfo=UTC), datetime(2022, 11, 26, 23, 15, tzinfo=UTC)], default='2022-11-26 23:15', timezone='UTC')
	verify('tz c', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 13:44', b'2020-02-22 12:44', b'x'], [datetime(2020, 9, 30, 11, 44, tzinfo=UTC), datetime(2020, 2, 22, 11, 44, tzinfo=UTC), datetime(2022, 11, 26, 23, 15, tzinfo=UTC)], default='2022-11-27 00:15', timezone=':Europe/Stockholm')

	# And now, for efficiency, all the remaining tests will be collected in
	# just three datasets to speed things up.
	# Each gets typed three times, as datetime, date and time.
	# They are also first tested with dsutil.strptime[_i](), both to test
	# those functions and to give better errors.

	from accelerator.dsutil import strptime, strptime_i

	# Test that the functions work at all, including default=
	assert strptime('2022-11-23', '%Y-%m-%d') == datetime(2022, 11, 23)
	assert strptime('abc', '%Y-%m-%d', default='nah') == 'nah'
	assert strptime_i('2022-11-23xyz', '%Y-%m-%d') == (datetime(2022, 11, 23), b'xyz')
	assert strptime_i('abc', '%Y-%m-%d', default='nah') == ('nah', b'abc')
	# If some matching succeeded the remaining string should reflect that even on failure.
	assert strptime_i('2022-xyz', '%Y-%m-%d', default='nah') == ('nah', b'xyz')

	tests = defaultdict(list)

	def good(want, *a):
		# datetime is a subclasses of date, so can't use isinstance
		if type(want) == date:
			want = datetime.combine(want, time(0, 0, 0))
		elif type(want) == time:
			want = datetime.combine(date(1970, 1, 1), want)
		for value in a:
			tests[pattern].append((value, want))
			got = strptime(value, pattern)
			assert got == want, "Parsing %r as %r\n    expected %s\n    got      %s" % (value, pattern, want, got,)
			value += 'x'
			got, remaining = strptime_i(value, pattern)
			assert got == want, "Parsing %r as %r\n    expected %s\n    got      %s" % (value, pattern, want, got,)
			assert remaining == b'x', "Parsing %r as %r left %r unparsed, expected %r" % (value, pattern, remaining, b'x',)

	def bad(*a):
		for value in a:
			tests[pattern].append((value, None))
			try:
				got = strptime(value, pattern)
				raise Exception("Parsing %r as %r gave %s, should have failed" % (value, pattern, got,))
			except ValueError:
				pass

	pattern = '%Y%m%d'
	#                         YYYYmmdd    YYYYmmdd    YYYYmmd    Ymmdd
	good(date(2019,  5, 21), '20190521', '2019 521')
	good(date(1970,  1,  1), '19700101', '1970 1 1', '1970 11')
	good(date(1945,  6, 20), '19450620')
	good(date(   1,  1,  1), '00010101', '   1 1 1', '  01 11', '1 101')
	good(date(9999, 12, 31), '99991231')
	#    YYYYmmd!    YYYYm!        YYYY!         YYYY!   !      Ymmdd!
	bad('19700100', '1970 01 01', '1970-01-01', '1980', 'nah', '1 1010')

	pattern = ' %Y %m %d '
	#                         YYYYmmdd
	good(date(2019,  5, 21), '20190521')
	#                            YYYY mm      dd    YYYY  m d
	good(date(1970,  1,  1), '   1970 01\n\n\n01', '1970\t1 1    ')
	#                         YYYY m dd    Y m d    Y    m      d
	good(date(1945,  6, 20), '1945 6 20')
	good(date(   1,  2,  3), '  01 2  3', '1 2 3', '1\t\r2\n\v\f3 \r ')
	#    YY m!     YYYY!
	bad('70 0 0', '1981')

	pattern = '%Yblah%m%d'
	#                         YYYYblahmmdd
	good(date(2019,  5, 21), '2019blah0521')
	good(date(1970,  1,  1), '1970blah0101')
	good(date(1945,  6, 20), '1945blah0620')
	good(date(  72, 11,  9), '  72blah11 9')
	#    YYYYblah!         YYYYblah!   !
	bad('1970blah-01-01', '1980blah', 'nah')

	pattern = '%Y%m%d%H%M%S'
	good(datetime(2013,  4,  9,  1,  2,  3),
		#YYYYmmddHHMMSS    YYYYmmddHHMMSS    YYYYmmddHHMMDD    YYYYmmddHHMMSS
		'2013 4 9 1 2 3', '2013 40901 203', '201304 901 2 3', '20130409010203',
	)
	#    YYYYmmd!           !                  YYYYmmd!           YYYYmmddHH!!
	bad('2013 4 09 1 2 3', ' 2013 4 9 1 2 3', '201304009010203', '20130409016003')

	pattern = '%H%M%y%m%d'
	#                                     HHMMyymmdd    HHMMyymmdd    HMMyymmd
	good(datetime(2019,  5, 21, 18, 52), '1852190521')
	good(datetime(1970,  1,  1,  0,  0), '0000700101')
	good(datetime(1978,  1,  1,  0,  0), '0000780101')
	good(datetime(2000,  3,  4,  5,  6), '0506000304', ' 5 6 0 3 4', '5 6 0 34')
	#    !        HHMMy!
	bad('today', '56034')

	pattern = '%Y%m%d %H%M%S.%f'
	#                                                 YYYYmmdd HHMMSS.fff
	good(datetime(2019,  5, 21, 18, 52,  6, 123000), '20190521 185206.123')
	#                                                 YYYYmmddHHMMSS.ffffff
	good(datetime(1970,  2,  3,  4,  5,  6,      7), '19700203040506.000007')
	#                                                 YYYYmmdd HHMMSS.f
	good(datetime(1978,  1,  1,  0,  0,  0,      0), '19780101 000000.0')
	#    YYYYMMDD  HHMMSS.!         !
	bad('19700203  040506.-00007', 'today')

	pattern = '%H:%M'
	#                   HH:MM    HH:MM    H:MM
	good(time( 3, 14), '03:14', ' 3:14', '3:14')
	good(time(18, 52), '18:52')
	good(time( 0, 42), '00:42')
	#    H!
	bad('25:10')

	pattern = '%H%M%%f%S' # this is HHMM followed by literal '%f' and then SS
	#                       HHMM%fSS    HHMM%fS    HMM%fSS
	good(time( 3, 14,  0), '0314%f00', ' 314%f0')
	good(time(18, 52,  9), '1852%f09', '1852%f9')
	good(time( 0, 42, 18), '0042%f18')
	good(time( 1,  2,  3), '0102%f03', ' 102%f3', '1 2%f 3')
	#    HHMM%fS!    HHMM%f!   HMM!        HMM%fS!
	bad('1938%f60', '1852%f', '1 02%f 3', '1 2%f ')

	# Test microseconds (%f)
	pattern = '%H%M %f.%S'
	#                               HHMMf.SS
	good(time( 3, 14,  0, 900000), '03149.00')
	#                               HHMMfff.SS
	good(time(18, 52,  9, 456000), '1852456.09')
	#                               HHMM   ffffff.SS
	good(time(19, 38, 44, 123456), '1938   123456.44')
	#                               HHMMffffff.SS
	good(time(18, 52,  9, 456000), '1852456   .09')
	#                               HHMMff.SS
	good(time( 0, 42, 18,      0), '004200.18')
	#    HHMMffffff.S!    HHMMffffff!
	bad('1938123456.60', '1852456    .09')

	# Just month and day (implicitly 1970)
	pattern = '%m%d'
	#                         mmdd    mmd    mmdd    mmd
	good(date(1970,  1,  2), '0102', '012', ' 102', ' 12')
	good(date(1970, 10, 20), '1020')
	good(date(1970,  6, 20), '0620')
	#           m!   mm!   mmd!   mmd!
	bad('nah', '1', '12', '12 ', '12  ')

	# Test that two digit years choose the right century
	pattern = '%y'
	#                         yy    y
	good(date(1969,  1,  1), '69')
	good(date(1970,  1,  1), '70')
	good(date(2000,  1,  1), '00', '0')
	good(date(2008,  1,  1), '08', '8')
	good(date(2019,  1,  1), '19')
	good(date(2068,  1,  1), '68')
	#    yy!     yy!    !
	bad('2000', '100', '-1')

	pattern = '%f%d'
	#                                                ffffffdd    ffd    ffdd
	good(datetime(1970, 1, 30, microsecond=300   ), '00030030')
	good(datetime(1970, 1,  6                    ), '00000006', '0 6', '0 06')
	good(datetime(1970, 1,  3, microsecond=30    ), '00003003')
	good(datetime(1970, 1, 11, microsecond=999999), '99999911')
	#    ffffffd!
	bad('99999999')

	pattern = '%f.%d'
	good(datetime(1970,  1, 30, microsecond=300000), '30.30')
	good(datetime(1970,  1,  6                    ), '0.06')
	good(datetime(1970,  1,  3, microsecond=300   ), '00030.03')
	good(datetime(1970,  1, 11, microsecond=999999), '999999.11')
	#    ffffff.d!
	bad('999999.99')

	pattern = '%f%d'
	#                                                 fffdd
	good(datetime(1970,  1, 30, microsecond=300000), '30 30')
	#                                                 ffdd    fffffdd    ffffffdd
	good(datetime(1970,  1,  6                    ), '0 06', '0    06', '0000   6')
	#                                                 ffffffdd    ffffffdd    fffffd
	good(datetime(1970,  1,  3, microsecond=300   ), '00030003', '00030 03', '0003 3')
	good(datetime(1970,  1, 11, microsecond=999999), '99999911')
	#    ffffffd!    ffffffdd!    ffffffd!
	bad('99999999', '999999 11', '0000    6')

	pattern = '%f%%f%%%d' # this is %f, literal '%f%', %d
	good(datetime(1970,  1, 30, microsecond=300000), '30%f%30')
	good(datetime(1970,  1,  6                    ), '0%f%06', '0000 %f%6')
	good(datetime(1970,  1,  3, microsecond=300   ), '00030%f%03')
	good(datetime(1970,  1, 11, microsecond=999999), '999999%f%11')
	#    ffffff%f%d!    !  (%f doesn't accept leading spaces)
	bad('999999%f%99', ' 30%f%30')

	# Avoid testing "far off" dates with functions that might use time_t
	# if long (the smallest type time_t is likely to be) is small.
	time_t_is_probably_big_enough = (len(struct.pack("@l", 0)) > 4)

	# "Unix" timestamps, seconds since 1970-01-01 00:00:00.
	# Uses gmtime_r, so limited by the platform time_t.
	pattern = '%s.%f'
	good(datetime(1970,  1,  1,  0,  0, 30, 300000), '30.30')
	good(datetime(2019,  5, 24,  1, 54, 13, 847211), '1558662853.847211')
	good(datetime(1970,  1,  1,  0,  0,  0, 100000), '0.1')
	if time_t_is_probably_big_enough:
		good(datetime(2286, 11, 20, 17, 46, 40, 234567), '10000000000.234567')
		# Since %f is separate this is actually later than -10000000000, which is unfortunate
		good(datetime(1653,  2, 10,  6, 13, 20, 111000), '-10000000000.111')
	#    !   !      !   !     !           !          !  (%s accepts no spaces)
	bad('', '.', '0.', '.1', ' 30.30', '30 .30', '30. 30')

	# "Java" timestamps, milliseconds since 1970-01-01 00:00:00.
	# Uses gmtime_r, so limited by the platform time_t.
	pattern = '%J'
	good(datetime(1970,  1,  1,  0,  0,  0,      0), '0')
	good(datetime(2019,  5, 24,  1, 54, 13, 847000), '1558662853847')
	good(datetime(1970,  1,  1,  0,  0,  0,   1000), '1')
	good(datetime(1969, 12, 31, 23, 59, 57, 995000), '-2005')
	if time_t_is_probably_big_enough:
		good(datetime(2286, 11, 20, 17, 46, 40, 234000), '10000000000234')
		# With %J the ms part is in the same number, and thus negative numbers are correctly handled.
		good(datetime(1653,  2, 10,  6, 13, 19, 889000), '-10000000000111')
	#    !    !      !    !  (%J accepts no spaces)
	bad('', '0x0', '0 ', ' 0')

	pattern = 'blah %Jbluh'
	good(datetime(1970,  1,  1,  0,  0,  0,      0), 'blah0bluh')
	good(datetime(1970,  1,  1,  0,  0, 30,      0), 'blah   30000bluh')
	good(datetime(1970,  1,  1,  0,  0,  0,   1000), 'blah1bluh')
	good(datetime(1969, 12, 31, 23, 59, 57, 995000), 'blah-2005bluh')
	bad('bla0bluh', 'blah0blu')

	# Excel dates. First in the default Lotus 1-2-3 format used by MS Excel.
	pattern = '%e'
	epoch = datetime(1899, 12, 31)
	if time_t_is_probably_big_enough:
		good(datetime(1900,  1,  1            ), '1')
		good(datetime(1900,  1,  2            ), '2')
		good(datetime(1900,  1,  2, 18        ), '2.75')
		good(datetime(1900,  2, 27            ), '58')
		good(datetime(1900,  2, 28            ), '59')
		good(datetime(1900,  3,  1            ), '60') # actually invalid (non-existant 1900-02-29)
		good(datetime(1900,  3,  1            ), '61') # same as 60
		good(datetime(1900,  3,  2            ), '62')
	assert datetime(1970, 1, 1) - epoch == timedelta(days=25568) # one off because of incorrect leap year 1900
	good(datetime(1970,  1,  1            ), '25569')
	if time_t_is_probably_big_enough:
		assert datetime(1800, 1, 2) - epoch == timedelta(days=-36522)
		good(datetime(1800,  1,  2            ), '-36522')
		good(datetime(1800,  1,  1, 18        ), '-36522.25')
		good(datetime(1800,  1,  1, 18        ), '4294930773.75') # gnumeric can get confused and produce this kind of thing
		good(datetime(1899, 12, 30            ), '-1')
		good(datetime(1899, 12, 30,  6        ), '-0.75')
		good(datetime(1899, 12, 31            ), '0')
		good(datetime(1899, 12, 31, 18        ), '0.75')
		assert datetime(9999, 12, 31) - epoch == timedelta(days=2958464) # one off because of incorrect leap year 1900
		good(datetime(9999, 12, 31            ), '2958465')
		good(datetime(9999, 12, 31, 23, 59, 58), '2958465.99998')
		assert datetime(1, 1, 1) - epoch == timedelta(days=-693594)
		good(datetime(   1,  1,  1            ), '-693594')
		good(datetime(   1,  1,  1,  0,  0,  1), '-693593.99999')
		good(datetime(1900,  1,  1, 23, 59, 59), '1.99999')
		good(datetime(1900,  1,  1, 23, 59, 58), '1.99998')
		good(datetime(1900,  1,  1, 23, 59, 57), '1.99997')
		good(datetime(1900,  1,  1, 23, 59, 57), '1.99996') # also 57!
		good(datetime(1900,  1,  1, 23, 59, 56), '1.99995')
	bad('-36522.-25')
	bad('2958466', '2958466.00001') # > positive max
	bad('-693595', '-693594.00001') # < negative max

	# Since 1 is the default flag for e, this is the same as %e
	pattern = '%1e'
	if time_t_is_probably_big_enough:
		good(datetime(1900,  1,  1            ), '1')
		good(datetime(1900,  2, 27            ), '58')
		good(datetime(1900,  2, 28            ), '59')
		good(datetime(1900,  3,  1            ), '60') # actually invalid (non-existant 1900-02-29)
		good(datetime(1900,  3,  1            ), '61') # same as 60
		good(datetime(1900,  3,  2            ), '62')
		good(datetime(1899, 12, 30            ), '-1')
	good(datetime(1970,  1,  1            ), '25569')
	# That's enough re-testing.

	# 0 is "Libre Office" dates, where 1900 is not a leap year.
	# (Dates before 1900-03-01 are offset compared to Lotus 1-2-3.)
	pattern = '%0e'
	epoch = datetime(1899, 12, 30)
	if time_t_is_probably_big_enough:
		good(datetime(1899, 12,  31           ), '1')
		good(datetime(1900,  1,  1            ), '2')
		good(datetime(1900,  1,  1, 18        ), '2.75')
		good(datetime(1900,  2, 26            ), '58')
		good(datetime(1900,  2, 27            ), '59')
		good(datetime(1900,  2, 28            ), '60')
		good(datetime(1900,  3,  1            ), '61')
		good(datetime(1900,  3,  2            ), '62')
	assert datetime(1970, 1, 1) - epoch == timedelta(days=25569)
	good(datetime(1970,  1,  1            ), '25569')
	if time_t_is_probably_big_enough:
		assert datetime(1800, 1, 1) - epoch == timedelta(days=-36522)
		good(datetime(1800,  1,  1            ), '-36522')
		good(datetime(1799, 12, 31, 18        ), '-36522.25')
		good(datetime(1799, 12, 31, 18        ), '4294930773.75') # gnumeric can get confused and produce this kind of thing
		good(datetime(1899, 12, 29            ), '-1')
		good(datetime(1899, 12, 29,  6        ), '-0.75')
		good(datetime(1899, 12, 30            ), '0')
		good(datetime(1899, 12, 30, 18        ), '0.75')
		assert datetime(9999, 12, 31) - epoch == timedelta(days=2958465)
		good(datetime(9999, 12, 31            ), '2958465')
		good(datetime(9999, 12, 31, 23, 59, 58), '2958465.99998')
		assert datetime(1, 1, 1) - epoch == timedelta(days=-693593)
		good(datetime(   1,  1,  1            ), '-693593')
		good(datetime(   1,  1,  1,  0,  0,  1), '-693592.99999')
		good(datetime(1899, 12 , 31, 23, 59, 59), '1.99999')
		good(datetime(1899, 12 , 31, 23, 59, 58), '1.99998')
		good(datetime(1899, 12 , 31, 23, 59, 57), '1.99997')
		good(datetime(1899, 12 , 31, 23, 59, 57), '1.99996') # also 57!
		good(datetime(1899, 12 , 31, 23, 59, 56), '1.99995')
	bad('-36522.-25')
	bad('2958466', '2958466.00001') # > positive max
	bad('-693594', '-693593.00001') # < negative max

	# 2 is "Mac Office" dates, epoch is 1904-01-01 and leap years are correct.
	pattern = '%2e'
	epoch = datetime(1904, 1, 1)
	good(datetime(1904,  1,  2            ), '1')
	good(datetime(1904,  1,  3            ), '2')
	good(datetime(1904,  1,  3, 18        ), '2.75')
	good(datetime(1904,  2, 28            ), '58')
	good(datetime(1904,  2, 29            ), '59')
	good(datetime(1904,  3,  1            ), '60')
	good(datetime(1904,  3,  2            ), '61')
	good(datetime(1904,  3,  3            ), '62')
	assert datetime(1974, 1, 2) - epoch == timedelta(days=25569)
	good(datetime(1974,  1,  2            ), '25569')
	if time_t_is_probably_big_enough:
		assert datetime(1804, 1, 3) - epoch == timedelta(days=-36522)
		good(datetime(1804,  1,  3            ), '-36522')
		good(datetime(1804,  1,  2, 18        ), '-36522.25')
		good(datetime(1804,  1,  2, 18        ), '4294930773.75') # gnumeric can get confused and produce this kind of thing
	good(datetime(1903, 12, 31            ), '-1')
	good(datetime(1903, 12, 31,  6        ), '-0.75')
	good(datetime(1904,  1,  1            ), '0')
	good(datetime(1904,  1,  1, 18        ), '0.75')
	if time_t_is_probably_big_enough:
		assert datetime(9999, 12, 31) - epoch == timedelta(days=2957003)
		good(datetime(9999, 12, 31            ), '2957003')
		good(datetime(9999, 12, 31, 23, 59, 58), '2957003.99998')
		assert datetime(1, 1, 1) - epoch == timedelta(days=-695055)
		good(datetime(   1,  1,  1            ), '-695055')
		good(datetime(   1,  1,  1,  0,  0,  1), '-695054.99999')
	good(datetime(1904,  1,  2, 23, 59, 59), '1.99999')
	good(datetime(1904,  1,  2, 23, 59, 58), '1.99998')
	good(datetime(1904,  1,  2, 23, 59, 57), '1.99997')
	good(datetime(1904,  1,  2, 23, 59, 57), '1.99996') # also 57!
	good(datetime(1904,  1,  2, 23, 59, 56), '1.99995')
	bad('-36522.-25')
	bad('2957004', '2957004.00001') # > positive max
	bad('-695056', '-695056.00001') # < negative max

	# Test wildcard characters.
	# '% ' whitespace (exactly one, plain space is any number)
	# '%.' any character except whitespace
	# '%*' any character including whitespace
	# '%#' any digit
	# '%@' any non-digit character, excluding whitespace
	# '%^' any non-digit character, including whitespace
	pattern = '%Y% %m%.%d%*%H%#%M%@%S%^'
	#                                         Y m.d*HH#M@S^    Y  mm.d*HH#M@S^
	good(datetime(   1,  2,  3,  4,  5,  6), '1 2x3x 495x6x', '1\t02x3x 495x6x')
	good(datetime(   1,  2,  3,  4,  5,  6), '1 2.3  495-6 ', '1\n02.3  495-6 ')
	#                                          ? ?    ? ? ?    Y m!
	bad(                                     '1.2.3  495-6 ', '1  02.3  495-6 ')
	bad(                                     '1 2 3  495-6 ')
	bad(                                     '1 2.3  4.5-6 ')
	bad(                                     '1 2.3  495 6 ')
	#                                         Y m.d*HH#M@SS!
	bad(                                     '1 2.3  495-111')
	#                                         Y m.d*H!
	bad(                                     '1 2.3 495-6 ')
	#                                         YYYY mm.dd*HH#M@S^
	good(datetime(   1,  2,  3,  4,  5,  6), '0001  22 33 445/6/')

	# Test separators with counts.
	# Only test with %@, as the repeat logic is shared.
	pattern = '%y%3@%m'
	good(date(2003, 6, 1), '3abc6', '3...6', '3+-_6', '3/@\\6')
	# Test with a range
	pattern = '%y%2,5@%m'
	good(date(2003, 6, 1), '3abcde6', '3..6', '3...6', '3....6')
	bad('3abcdef6', '3.6')
	# Test with a range that includes 0
	pattern = '%y%0,2@%m'
	good(date(2003, 6, 1), '03ab6', '03a6', '036')
	bad('36', '3abc6')

	# Let's generate an exhaustive search (within ASCII) for all separators
	# (and no \0, strptime doesn't handle it.)
	def negate_ascii(chars):
		return {c for c in map(chr, range(1, 128)) if c not in chars}
	for sep, allowed in (
		('% ', '\t\n\v\f\r '),
		('%.', negate_ascii('\t\n\v\f\r ')),
		('%*', negate_ascii('')),
		('%#', '0123456789'),
		('%@', negate_ascii('0123456789\t\n\v\f\r ')),
		('%^', negate_ascii('0123456789')),
	):
		pattern = '%Y' + sep + '%m'
		for c in allowed:
			good(date(3456, 12, 1), '3456' + c + '12')
			bad('3456' + c + c + '12') # don't match several
		for c in negate_ascii(allowed):
			bad('3456' + c + '12')

	# Test optional tails. %/ makes the rest of the pattern optional,
	# but the whole string still has to match.
	pattern = '%d%/%m%y'
	#                             ddmmyy    ddmm    ddm    dd    d
	good(datetime(1970,  1,  6), '060170', '0601', '061', '06', '6')
	good(datetime(1970,  2,  3), '030270', '0302', '032')
	#    ddmm!    d!
	bad('0601x', '60170')

	# Java timestamp with optional ms-fraction
	# (which overrides the fractional part of the full timestamp when used)
	pattern = '%J%/.%3f'
	#                                                JJJJJ    JJJJJ.    JJJJJ.FFF    JJJJJ.FFF
	good(datetime(1970,  1,  1,  0,  0, 30       ), '30000', '30000.', '30000.000', '30303.0  ')
	good(datetime(1970,  1,  1,  0,  0, 30, 50000), '30050', '30050.', '30000.050', '30303.05 ')
	#    JJJJJ!    JJJJJ.FFF!    !
	bad('30050 ', '30000.05  ', ' 30050')

	# Test %?, making the next field(s) optional.
	# The way this works is that as long as it (all) matches it is used,
	# there is no backtracking if fields after the optional part don't match.

	# Test single optional values
	pattern = '%Y%?%b%d'
	#                             YYYYbbbd    YYYYbbbbbbbdd    YYYYd
	good(datetime(2023,  1,  2), '2023jan2', '2023january02', '20232')
	good(datetime(2023,  1,  2), '2023jAN2', '2023JanuarY02', '202302')
	#    YYYY!   YYYYbbb!   YYYYbb! (and "jab2" isn't ok for d)
	bad('2023', '2023jan', '2023jab2')

	# Test optional values with a count (%H only applies if %b worked)
	pattern = '%Y%2?%b%H%d'
	#                                 YYYYbbbHHd    YYYYbbbbbbbHHdd
	good(datetime(2023,  1,  3, 16), '2023jan163', '2023january1603')
	#                                 YYYYdd
	good(datetime(2023,  1, 16,  0), '202316')
	#    YYYYdd!    YYYYbbbbbbbHH!
	bad('2023163', '2023january03')

	# %? doesn't have to match %-things, here it's also matching the ":"
	pattern = '%Y-%m %3?%H:%M %d'
	#                                     YYYY-mm HH:MM dd    YYYY-mm HH:MMd
	good(datetime(2018, 10,  1, 12,  1), '2018-10 12:01 01', '2018-10 12: 11')
	#                                     YYYY-mm dd    YYYY-mm HH:MM dd
	good(datetime(2018, 10, 11,  0,  0), '2018-10 11', '2018-10 00:00 11')
	#    YYYY-mm HH:MM!   YYYY-MM HH:MM!   YYYY-MM HH:!! (fails '  ' as MM, then fails with trailing ':  1' after the dd)
	bad('2018-10 00:00', '2018-10 00: 1', '2018-10 05:  1')

	# Make sure the spaces weren't important (except for readability)
	pattern = '%Y-%m%3?%H:%M%d'
	#                                     YYYY-mmHH:MMdd    YYYY-mmHH:MMd
	good(datetime(2018, 10,  1, 12,  1), '2018-1012:0101', '2018-1012: 11')
	#                                     YYYY-mmdd    YYYY-mmHH:MMdd
	good(datetime(2018, 10, 11,  0,  0), '2018-1011', '2018-1000:0011')
	#    YYYY-mmHH:MM!   YYYY-MMHH:MM!   YYYY-MMHH:!! (fails '  ' as MM, then fails with trailing ':  1' after the dd)
	bad('2018-1000:00', '2018-1000: 1', '2018-1005:  1')

	# Test optional values with a range (%H only applies if %b worked, and %M only if %H)
	pattern = '%d%2,3?%b%H%M%Y'
	#                                     ddbbbHHMMYYYY    dbbbHHMMYYYY
	good(datetime(2023,  1,  3, 16,  4), '03jan16042023', '3jan16 42023')
	#                                     ddbbbHHYYYY  (98 is not a possible M value)
	good(datetime(9876,  1,  3, 16    ), '03jan169876')
	#                                     ddYYYY
	good(datetime(1604,  1,  3        ), '031604')
	#    ddYYYY!       ddbbbH!  (99 is not a possible H value, so it fails there, and then on 'j' for Y)
	bad('0316042023', '03jan9999')

	# Recursive optional bits
	# This is hard to read, maybe "(a(b%Y)c)(d(efg)h)(%Y)-%m (ABC[D%H%M]) %d" helps?
	# A %? inside a %? counts as a single token, regardless of count.
	pattern = '%3?a%2?b%Yc%3?d%3?efgh%?%Y-%m %3,6?ABCD%H%M %d'
	#                                  abYYYYcdefghYYYY-mm ABCDHHMM d
	good(datetime(2022, 11, 7, 1, 2), 'ab1998cdefgh2022-11 ABCD0102 7',)
	#                                  abYYYYc-mm d    abYYYYc-mm ABCd
	good(datetime(2023, 12, 8, 0, 0), 'ab2023c-12 8', 'ab2023c-12 ABC8')
	#    abYYYYcdefghYYYY-mm !            abYYYYcdefghYYYY-mm AB!
	bad('ab1998cdefgh2022-11 BCD0102 7', 'ab1998cdefgh2022-11 AB 7')

	# Test the else function %:
	# Accept YYYYmmdd if that fits, otherwise an Excel date
	pattern = '%3?%Y%m%d%:%e'
	#                       eeeee    YYYYmmdd
	good(date(1997, 5, 5), '35555', '19970505')
	bad('')

	# A plausable date thing, accepting both name and number for the month.
	# Try it in both orders.
	for pattern in ('%d-%?%b%:%m-%y', '%d-%?%m%:%b-%y'):
		#                         dd-mm-yy,   dd-m-yy,   dd-bbb-yy,   dd-bbbbb-yy
		good(date(1997,  3, 31), '31-03-97', '31-3-97', '31-mar-97', '31-march-97')
		#    dd-bbb!        dd-bbb!       dd-bbb-yy!      dd-mm!         dd-m!
		bad('31-mar03-97', '31-mar3-97', '31-mar-03-97', '31-03mar-97', '31-3mar-97', '31--97')

	# Test %: recursively too.
	# This is either the same as above or the same in opposite order.
	# I.e. it's one of those terrible things that guesses which end the
	# year is in depending on if the first number if valid as a day or not.
	#             11233334444566   11233334444566
	pattern = '%6?%d-%?%b%:%m-%y%6:%y-%?%b%:%m-%d'
	good(date(1985, 12, 24), '24-dec-85', '24-12-85', '85-12-24', '85-dec-24')
	#        both years,  repeating everything in various orders
	bad('', '85-dec-85', '24-12-8585-12-24', '24-12-8524-12-85', '85-12-2485-12-24', '85-12-2424-12-85')

	# Combining %/ with %? and %:

	# If there is a month name %/ starts applying, thus making the X optional,
	# but the X still has to match to make the %3? happy!
	# So in practice the %/ only applies to the -%d, and of course only
	# if the %b matches.
	pattern = '%y-%3?%b%/X%:%m-%d'
	#                         yy-mm-dd    yy-bbbX-dd    yy-bbbX-    yy-bbbX
	good(date(1978, 11,  1), '78-11-01', '78-novX-01', '78-novX-', '78-novX')
	good(date(2008,  4, 14), '08-04-14', '08-aprX-14')
	#    yy-mm!   yy-bbb!      yy-bbb!    yy-bbb!
	bad('78-11', '78-nov-01', '78-nov-', '78-nov')

	# Same thing, but put the %/ in the else.
	# Does not match exactly the same things as above, because %: only
	# cares about the count for skipping when the initial %? matched. In
	# other words, the X does not have to match to make the %/ apply.
	pattern = '%y-%?%m%3:%b%/X-%d'
	#                         yy-mm-dd    yy-bbbX-dd    yy-bbbX-    yy-bbbX    yy-bbb
	good(date(1978, 11,  1), '78-11-01', '78-novX-01', '78-novX-', '78-novX', '78-nov')
	good(date(2008,  4, 14), '08-04-14', '08-aprX-14')
	#    yy-mm!   yy-bbb!      yy-bbb!
	bad('78-11', '78-nov-01', '78-nov-')

	# Use %:%3? instead of %3: to almost match the same as in the first case,
	# but not quite as that then makes both sides optional.
	pattern = '%y-%?%m%:%3?%b%/X-%d'
	#                         yy-mm-dd    yy-bbbX-dd    yy-bbbX-    yy-bbbX
	good(date(1978, 11,  1), '78-11-01', '78-novX-01', '78-novX-', '78-novX')
	good(date(2008,  4, 14), '08-04-14', '08-aprX-14')
	#                         yy--dd
	good(date(1978,  1, 23), '78--23')
	#    yy-mm!   yy-bbb!      yy-bbb!    yy-bbb!
	bad('78-11', '78-nov-01', '78-nov-', '78-nov')

	# Test %I and %p.
	pattern = '%I%p'
	good(time( 0,  0), '12am')
	good(time( 1,  0), '01Am', '1aM')
	good(time(12,  0), '12pm')
	good(time(13,  0), '01Pm', '1pM')
	bad('0pm', '00pm', '13pm', '12', '13')

	# In "wrong" order.
	pattern = '%p%I'
	good(time( 0,  0), 'AM12')
	good(time( 1,  0), 'AM01', 'AM1')
	good(time(12,  0), 'PM12')
	good(time(13,  0), 'PM01', 'PM1')
	bad('PM0', 'PM00', 'PM13', '12', 'PM')

	# Test that %p works with %H. (Has to come after of course.)
	pattern = '%H%p'
	good(time( 2,  0), '14AM', '02AM', '2AM')
	good(time(14,  0), '14PM', '02PM', '2PM')
	bad('14')

	# Test that %p does not affect %H when %p comes first
	pattern = '%p%H'
	good(time( 2,  0), 'AM02', 'PM02')
	good(time(14,  0), 'AM14', 'PM14')

	# Test %C
	pattern = '%C%y'
	good(date(   1, 1, 1), '0001')
	good(date(1850, 1, 1), '1850')
	good(date(2070, 1, 1), '2070')
	good(date(9999, 1, 1), '9999')

	# In "wrong" order
	pattern = '%y%C'
	good(date(   1, 1, 1), '0100')
	good(date(1850, 1, 1), '5018')
	good(date(2070, 1, 1), '7020')
	good(date(9999, 1, 1), '9999')

	# Test %C to override century from %Y
	pattern = '%Y%C'
	good(date(   1, 1, 1), '200100')
	good(date(1850, 1, 1), '195018')
	good(date(2070, 1, 1), '187020')
	good(date(9999, 1, 1), '199999')

	# Test %Y to override %C
	pattern = '%C%Y'
	good(date(   1, 1, 1), '200001')
	good(date(1850, 1, 1), '191850')
	good(date(2070, 1, 1), '182070')
	good(date(9999, 1, 1), '199999')

	# Test keeping century from %Y with %y
	pattern = '%Y%y'
	good(date(   1, 1, 1), '008801')
	good(date(1850, 1, 1), '186850')
	good(date(2070, 1, 1), '202370')
	good(date(9999, 1, 1), '992299')

	# A complete and totally idiotic format.
	pattern = '%y%I%M%C%d%p%b%S'
	#                                         yyIIMMCCDppbbbS    yyIIMMCCddppbbbbbbbSS
	good(datetime(1956, 10,  5, 23, 34,  8), '561134195pmoct8', '56113419 5Pmoctober08')
	#    yy!!MMCCDppbbbS    yyIIMMCCddppbbbbbb!     yyIIMMCCdppbbbS!
	bad('562134195pmoct8', '56113419 5PMoctobe08', '561134195pmoct8 ')

	# Save all of these in three datasets with one column per pattern.
	#     One with only good values
	#     One with only good values, with trailing garbage (typed with *i:)
	#     One with both good and bad values (typed with defaults)
	# Columns are padded with None-values so they are all the same length.
	# The order of each column is random, just in case there is some
	# inter-value state.
	# Each dataset is typed as datetime, date and time.
	def one(name, type_suffix, value_suffix, with_defaults):
		dw = DatasetWriter(name=name, allow_missing_slices=True)
		colnames = []
		to_write = []
		want = []
		max_len = max(len(values) for _, values in tests.items())
		for pattern, values in tests.items():
			dw.add(pattern, 'ascii', none_support=True)
			colnames.append(pattern)
			if not with_defaults: # remove bad
				values = [v for v in values if v[1] is not None]
			values = values + [(None, None)] * (max_len - len(values))
			random.shuffle(values)
			to_write.append([None if w is None else w + value_suffix for w, _ in values])
			want.append([w for _, w in values])
		dw.set_slice(0)
		for line in zip(*to_write):
			dw.write_dict(dict(zip(colnames, line)))
		source = dw.finish()

		def check(type_as, fix):
			column2type = {col: type_as + type_suffix + ':' + col for col in colnames}
			kw = dict(source=source, column2type=column2type)
			if with_defaults:
				kw['defaults'] = {col: None for col in colnames}
			ds = subjobs.build('dataset_type', **kw).dataset()
			for col, wrote, good in zip(colnames, to_write, want):
				got = list(ds.iterate(0, col))
				assert len(got) == len(wrote) == len(good) == max_len, col
				for got, wrote, good in zip(got, wrote, good):
					if good is not None:
						good = fix(good)
					assert got == good, 'Typing %r as %r gave %r, expected %r' % (wrote, column2type[col], got, good,)

		check('datetime', lambda dt: dt)
		check('date', lambda dt: dt.date())
		check('time', lambda dt: dt.time())

	one('good datetimes', '', '', False)
	one('good datetimes with trailing garbage', 'i', 'whee', False)
	one('good and bad datetimes', '', '', True)

	if sys.version_info >= (3, 6):
		# These can't be part of the dataset_type calls above, as they will only
		# fail when typed as dates, not as times. Therefore, they are after the
		# one() calls, so only the strptime part of the testing is done.
		pattern = '%Y-%m-%d'
		good(date(1992, 2, 29), '1992-02-29')
		bad('2019-02-29', '1970-02-31', '1980-06-31')
		# And then use verify() to test the same values in dataset_type.
		for type_as, func in (
			('date', date),
			('datetime', datetime),
		):
			verify('nearly good %s YYYY-MM-DD' % (type_as,), ['%s:%s' % (type_as, pattern,)], [b'2019-02-29', b'1970-02-31', b'1980-06-31', b'1992-02-29'], [None, None, None, func(1992, 2, 29)], None, False)


def test_filter_bad_across_types():
	columns={
		'bytes': 'bytes',
		'float64': 'bytes',
		'int32_10': 'ascii',
		'json': 'unicode',
		'number:int': 'unicode',
		'unicode:utf-8': 'bytes',
	}
	# all_good, *values
	# Make sure all those types (except bytes) can filter other lines,
	# and be filtered by other lines. And that several filtering values
	# is not a problem (line 11).
	data = [
		[True,  b'first',    b'1.1', '1',  '"a"',   '001', b'ett',],
		[True,  b'second',   b'2.2', '2',  '"b"',   '02',  b'tv\xc3\xa5',],
		[True,  b'third',    b'3.3', '3',  '["c"]', '3.0', b'tre',],
		[False, b'fourth',   b'4.4', '4',  '"d"',   '4.4', b'fyra',],       # number:int bad
		[False, b'fifth',    b'5.5', '-',  '"e"',   '5',   b'fem',],        # int32_10 bad
		[False, b'sixth',    b'6.b', '6',  '"f"',   '6',   b'sex',],        # float64 bad
		[False, b'seventh',  b'7.7', '7',  '{"g"}', '7',   b'sju',],        # json bad
		[False, b'eigth',    b'8.8', '8',  '"h"',   '8',   b'\xa5\xc3tta',],# unicode:utf-8 bad
		[True,  b'ninth',    b'9.9', '9',  '"i"',   '9',   b'nio',],
		[True,  b'tenth',    b'10',  '10', '"j"',   '10',  b'tio',],
		[False, b'eleventh', b'11a', '1-', '"k",',  '1,',  b'elva',],       # float64, int32_10 and number:int bad
		[True,  b'twelfth',  b'12',  '12', '"l"',   '12',  b'tolv',],
	]
	want_bad = [tuple(l[1:]) for l in data if not l[0]]
	dw = DatasetWriter(name="filter bad across types", columns=columns, allow_missing_slices=True)
	cols_to_check = ['int32_10', 'bytes', 'json', 'unicode:utf-8']
	if PY3:
		# z so it sorts last.
		dw.add('zpickle', 'pickle')
		cols_to_check.append('zpickle')
		for ix in range(len(data)):
			data[ix].append({ix})
	dw.set_slice(0)
	want = []
	def add_want(ix):
		v = data[ix]
		want.append((int(v[3]), v[1], json.loads(v[4]), v[6].decode('utf-8'),))
		if PY3:
			want[-1] = want[-1] + (v[7],)
	for ix, v in enumerate(data):
		if v[0]:
			add_want(ix)
		dw.write(*v[1:])
	source_ds = dw.finish()
	# Once with just filter_bad, once with some defaults too.
	defaults = {}
	for _ in range(2):
		jid = subjobs.build(
			'dataset_type',
			datasets=dict(source=source_ds),
			options=dict(column2type={t: t for t in columns}, filter_bad=True, defaults=defaults),
		)
		typed_ds = Dataset(jid)
		got = list(typed_ds.iterate(0, cols_to_check))
		assert got == want, "Expected %r, got %r from %s (from %r%s)" % (want, got, typed_ds, source_ds, ' with defaults' if defaults else '')
		bad_ds = Dataset(jid, 'bad')
		got_bad = list(bad_ds.iterate(0, sorted(columns)))
		assert got_bad == want_bad, "Expected %r, got %r from %s (from %r%s)" % (want_bad, got_bad, bad_ds, source_ds, ' with defaults' if defaults else '')
		# make more lines "ok" for the second lap
		if not defaults:
			want_bad.pop(0) # number:int
			want_bad.pop(1) # float64
			want_bad.pop(1) # json
		defaults = {'number:int': '0', 'float64': '0', 'json': '"replacement"'}
		add_want(3)
		add_want(5)
		data[6][4] = '"replacement"'
		add_want(6)
		want.sort() # adding them out of order, int32_10 sorts correctly.

def test_rename():
	def mk(name, colnames="abc", **kw):
		dw = DatasetWriter(name='rename_' + name, **kw)
		assert len(colnames) == 3
		for name, typ in zip(colnames, ('ascii', 'bytes', 'unicode')):
			dw.add(name, typ, none_support=(typ == 'unicode'))
		dw.get_split_write()('0', b'1', '2')
		return dw.finish()
	plain = mk('plain')
	with_hashlabel = mk('with_hashlabel', hashlabel='a')

	# Rename all, with some name hidden. Untyped a remains visible.
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(b='int32_10', c='int64_10', d='number'),
		rename=dict(a='b', b='c', c='d'),
		source=plain,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type, col.none_support) for name, col in typed_ds.columns.items())
	assert coltypes == [('a', 'ascii', False), ('b', 'int32', False), ('c', 'int64', False), ('d', 'number', True)], coltypes
	assert list(typed_ds.iterate(None)) == [('0', 0, 1, 2)]

	# Rename hashlabel a => b, c untouched except for rehashing
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(b='int32_10'),
		rename=dict(a='b'),
		source=with_hashlabel,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type, col.none_support) for name, col in typed_ds.columns.items())
	assert coltypes == [('b', 'int32', False), ('c', 'unicode', True)], coltypes
	assert list(typed_ds.iterate(None)) == [(0, '2')]
	assert typed_ds.hashlabel == 'b'

	# Discard hashlabel
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(b='int32_10'),
		rename=dict(a=None),
		source=with_hashlabel,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type, col.none_support) for name, col in typed_ds.columns.items())
	assert coltypes == [('b', 'int32', False), ('c', 'unicode', True)], coltypes
	assert list(typed_ds.iterate(None)) == [(1, '2')]
	assert typed_ds.hashlabel == None

	# test renaming several columns to the same name
	# (but in different parts of the chain)
	plain2 = mk('plain2', colnames="def", previous=plain)
	plain3 = mk('plain3', colnames="abi", previous=plain2)
	# once without rehashing, and once with
	hash_slice = typed_writer('number').hash(1) % g.slices
	for hashlabel, want_slice in ((None, 0), ('one', hash_slice)):
		jid = subjobs.build(
			'dataset_type',
			column2type=dict(zero='number', one='number', two='number'),
			hashlabel=hashlabel,
			rename=dict(a='zero', d='zero', b='one', e='one', c='two', f='two', i='two'),
			source=plain3,
		)
		typed_dss = jid.dataset().chain()
		assert len(typed_dss) == 3
		for org_cols, ds in zip(["abc", "def", "abi"], typed_dss):
			assert ds.hashlabel == hashlabel
			assert set(ds.columns) == set(org_cols) | {'zero', 'one', 'two'}
			assert list(ds.iterate(want_slice)) == [('0', b'1', '2', 1, 2, 0)]
			assert sum(ds.lines) == 1

def test_filter_bad_with_rename_and_chain():
	dw = DatasetWriter(name="filter bad with rename", allow_missing_slices=True)
	dw.add('a', 'ascii')
	dw.add('b', 'bytes')
	dw.add('c', 'unicode')
	dw.set_slice(0)
	dw.write('0', b'1', '2')
	dw.write('9', B'A', 'B')
	dw.write('C', B'D', 'E')
	source_ds = dw.finish()
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(b='int32_10', c='int64_16', d='int32_16'),
		filter_bad=True,
		rename=dict(a='b', b='c', c='d'),
		source=source_ds,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type) for name, col in typed_ds.columns.items())
	assert coltypes == [('a', 'ascii'), ('b', 'int32'), ('c', 'int64'), ('d', 'int32')], coltypes
	assert list(typed_ds.iterate(0)) == [('0', 0, 1, 2), ('9', 9, 10, 11)]
	bad_ds = jid.dataset('bad')
	coltypes = sorted((name, col.type) for name, col in bad_ds.columns.items())
	assert coltypes == [('b', 'ascii'), ('c', 'bytes'), ('d', 'unicode')], coltypes
	assert list(bad_ds.iterate(0)) == [('C', b'D', 'E')]

	dw = DatasetWriter(name="filter bad with rename chain", allow_missing_slices=True, previous=source_ds)
	dw.add('a', 'ascii')
	dw.add('b', 'ascii')
	dw.add('c', 'ascii')
	dw.set_slice(0)
	dw.write('3', '4', '5')
	dw.write('6', '7', 'eight')
	source_ds = dw.finish()
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(a='number', b='int32_10', c='int64_10'),
		defaults=dict(a='8'),
		filter_bad=True,
		rename=dict(a='b', b='c', c='a'),
		source=source_ds,
	)
	typed_ds = jid.dataset()
	assert len(typed_ds.chain()) == 2
	coltypes = sorted((name, col.type) for name, col in typed_ds.columns.items())
	assert coltypes == [('a', 'number'), ('b', 'int32'), ('c', 'int64')], coltypes
	assert list(typed_ds.iterate_chain(0)) == [(2, 0, 1), (5, 3, 4), (8, 6, 7)]
	bad_ds = jid.dataset('bad')
	assert len(bad_ds.chain()) == 2
	coltypes = sorted((name, col.type) for name, col in bad_ds.columns.items())
	assert coltypes == [('a', 'ascii'), ('b', 'ascii'), ('c', 'ascii')], coltypes
	coltypes = sorted((name, col.type) for name, col in bad_ds.previous.columns.items())
	assert coltypes == [('a', 'unicode'), ('b', 'ascii'), ('c', 'bytes')], coltypes
	assert list(bad_ds.iterate_chain(0)) == [('B', '9', b'A'), ('E', 'C', b'D')]

def test_None():
	types = {
		'bool': 'floatbool',
		'complex32': 'complex32', 'complex64': 'complex64',
		'float32': 'float32', 'float64': 'float64',
		'int32': 'int32_10', 'int64': 'int64_10',
		'number': 'number',
		'date': 'date:%Y-%m-%d',
		'datetime': 'datetime:%Y-%m-%dT%H:%M:%S',
		'time': 'time:%H:%M:%S',
		'ascii': 'ascii:strict', 'unicode': 'unicode:utf-8',
	}
	values = {
		'bool': b'-1',
		'date': b'2022-07-08',
		'datetime': b'2022-07-08T22:23:24',
		'time': b'22:23:24',
		'ascii': b'text',
		'unicode': b'\xe5\xad\x97',
	}
	values = {k: values.get(k, b'%d' % (ix,)) for ix, k in enumerate(sorted(types))}
	want_values = {
		'bool': True,
		'date': date(2022, 7, 8),
		'datetime': datetime(2022, 7, 8, 22, 23, 24),
		'time': time(22, 23, 24),
		'ascii': 'text',
		'unicode': '字',
	}
	want_values = {k: want_values.get(k, ix) for ix, k in enumerate(sorted(types))}
	def test(name, dw_kw, type_kw, write_lines, want_none_count):
		dw = DatasetWriter(name=name, allow_missing_slices=True, **dw_kw)
		for typ in types:
			dw.add(typ, 'bytes', none_support=True)
		dw.set_slice(0)
		for d in write_lines:
			dw.write(**d)
		source_ds = dw.finish()
		ds = subjobs.build(
			'dataset_type',
			source=source_ds,
			column2type=types,
			**type_kw
		).dataset()
		for colname, value in want_values.items():
			want = [value] + [None] * want_none_count
			assert list(ds.iterate_chain(0, colname)) == want, colname
		return source_ds
	# One line with good values, one with Nones
	prev = test('None values', {}, {}, [values, dict.fromkeys(values, None)], 1)
	# One line with bad values (plus the previous ds)
	test('bad None values', {'previous': prev}, {'defaults': dict.fromkeys(values, None)}, [dict.fromkeys(values, b'\xff')], 2)

def test_column_discarding():
	dw = DatasetWriter(name='column discarding')
	dw.add('a', 'bytes')
	dw.add('b', 'bytes')
	dw.add('c', 'bytes')
	w = dw.get_split_write()
	w(b'a', b'b', b'c')
	source = dw.finish()

	# Discard b because it's not typed
	ac_implicit = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		discard_untyped=True,
	).dataset()
	assert sorted(ac_implicit.columns) == ['a', 'c'], '%s: %r' % (ac_implicit, sorted(ac_implicit.columns),)
	assert list(ac_implicit.iterate(None)) == [('a', 'c',)], ac_implicit

	# Discard b explicitly
	ac_explicit = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		rename=dict(b=None),
	).dataset()
	assert sorted(ac_explicit.columns) == ['a', 'c'], '%s: %r' % (ac_explicit, sorted(ac_explicit.columns),)
	assert list(ac_explicit.iterate(None)) == [('a', 'c',)], ac_explicit

	# Discard c by overwriting it with b. Keep untyped b.
	ac_bASc = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		rename=dict(b='c'),
	).dataset()
	assert sorted(ac_bASc.columns) == ['a', 'b', 'c'], '%s: %r' % (ac_bASc, sorted(ac_bASc.columns),)
	assert list(ac_bASc.iterate(None)) == [('a', b'b', 'b',)], ac_bASc

	# Discard c by overwriting it with b. Also type b as a different type.
	abc_bASc = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', b='strbool', c='ascii'),
		rename=dict(b='c'),
	).dataset()
	assert sorted(abc_bASc.columns) == ['a', 'b', 'c'], '%s: %r' % (abc_bASc, sorted(abc_bASc.columns),)
	assert list(abc_bASc.iterate(None)) == [('a', True, 'b',)], abc_bASc

def test_rehash_with_empty_slices():
	dw = DatasetWriter(name='rehash with empty slices', hashlabel='a')
	dw.add('a', 'ascii')
	dw.add('b', 'ascii')
	w = dw.get_split_write()
	w('a', '42')
	w('42', 'b')
	source = dw.finish()
	hashfunc = typed_writer('int32').hash
	def verify_hashing(caption, want_values, **kw):
		ds = subjobs.build(
			'dataset_type',
			source=source,
			column2type=dict(a='int32_10'),
			caption=caption,
			**kw
		).dataset()
		got_values = set()
		for sliceno in range(g.slices):
			for got in ds.iterate(sliceno):
				assert hashfunc(got[0]) % g.slices == sliceno
				assert got not in got_values
				got_values.add(got)
		assert want_values == got_values
	verify_hashing('with discard', {(42, 'b',)}, filter_bad=True)
	# using defaults uses some different code paths
	verify_hashing('with default=0 (probably two slices)', {(0, '42',), (42, 'b',)}, defaults=dict(a='0'))
	verify_hashing('with default=42 (one slice)', {(42, '42',), (42, 'b',)}, defaults=dict(a='42'))

def test_varying_hashlabel():
	columns = {'a': 'ascii', 'b': 'ascii', 'c': 'ascii'}
	def mk(name, hashlabel, previous):
		dw = DatasetWriter(name=name, hashlabel=hashlabel, previous=previous, columns=columns)
		w = dw.get_split_write()
		w('1', '2', '3')
		return dw.finish()
	ds1 = mk('hlchain0', None, None)
	ds2 = mk('hlchain1', 'a', ds1)
	ds3 = mk('hlchain2', 'b', ds2)
	ds4 = mk('hlchain3', 'c', ds3)
	typed_dss = subjobs.build(
		'dataset_type',
		source=ds4,
		column2type=dict(a='int32_10', b='int32_10', c='int32_10'),
	).dataset().chain()
	assert list(typed_dss.iterate(None)) == 4 * [(1, 2, 3)]
	assert [ds.hashlabel for ds in typed_dss] == [None, 'a', 'b', 'c']
	h = typed_writer('int32').hash
	assert typed_dss[0].lines[0] == 1
	for ix in (1, 2, 3):
		assert typed_dss[ix].lines[h(ix) % g.slices] == 1


def synthesis():
	test_bytes()
	test_ascii()
	test_unicode()
	test_numbers()
	test_datetimes()
	test_None()
	test_column_discarding()
	test_rehash_with_empty_slices()
	test_rename()
	test_varying_hashlabel()

	verify('json', ['json'],
		[b'null', b'[42, {"a": "b"}]', b'\r  {  "foo":\r"bar" \r   }\t ', b'nope'],
		[None, [42, {'a': 'b'}], {'foo': 'bar'}, ['nah']],
		default='["nah"]', all_source_types=True,
	)

	test_filter_bad_across_types()
	test_filter_bad_with_rename_and_chain()

	for t in (all_typenames - used_typenames):
		print(t)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2024 Carl Drougge                                     #
# Modifications copyright (c) 2020-2021 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
import errno
import os
from os.path import dirname, basename, realpath, join, expanduser
import locale
from collections import namedtuple
from glob import glob
import re
import shlex
import signal
from argparse import RawDescriptionHelpFormatter

from accelerator.colourwrapper import colour
from accelerator.error import UserError
from accelerator.shell.parser import ArgumentParser

cfg = None
user_cfg = None

def find_cfgs(basedir='.', wildcard=False):
	"""Find all accelerator.conf (or accelerator*.conf if wildcard=True)
	starting at basedir and continuing all the way to /, yielding them
	from the deepest directory first, starting with accelerator.conf (if
	present) and then the rest in sorted order."""

	cfgname = 'accelerator.conf'
	if wildcard:
		pattern = 'accelerator*.conf'
	else:
		pattern = cfgname
	orgdir = os.getcwd()
	basedir = realpath(basedir)
	while basedir != '/':
		try:
			os.chdir(basedir)
			fns = sorted(glob(pattern))
		finally:
			os.chdir(orgdir)
		if cfgname in fns:
			fns.remove(cfgname)
			fns.insert(0, cfgname)
		for fn in fns:
			yield join(basedir, fn)
		basedir = dirname(basedir)

def load_some_cfg(basedir='.', all=False):
	global cfg

	basedir = realpath(basedir)
	cfgs = find_cfgs(basedir, wildcard=all)
	if all:
		found_any = False
		# Start at the root, so closer cfgs override those further away.
		for fn in reversed(list(cfgs)):
			try:
				load_cfg(fn)
				found_any = True
			except Exception:
				# As long as we find at least one we're happy.
				pass
		if not found_any:
			raise UserError("Could not find 'accelerator*.conf' in %r or any of its parents." % (basedir,))
		cfg.config_filename = None
	else:
		try:
			fn = next(cfgs)
		except StopIteration:
			raise UserError("Could not find 'accelerator.conf' in %r or any of its parents." % (basedir,))
		load_cfg(fn)

def load_cfg(fn):
	global cfg

	from accelerator.configfile import load_config
	from accelerator.job import WORKDIRS

	cfg = load_config(fn)
	for k, v in cfg.workdirs.items():
		if WORKDIRS.get(k, v) != v:
			print("WARNING: %s overrides workdir %s" % (fn, k,), file=sys.stderr)
		WORKDIRS[k] = v
	return cfg

def unpath(path):
	while path in sys.path:
		sys.path.pop(sys.path.index(path))

# This is used as "cfg" when no config is found (or parsing fails).
# Once the command tries to use cfg the original exception from loading the
# config is raised. But if the command doesn't use the cfg no exception is
# raised, so e.g. --help works.
class NoConfig(object):
	def __init__(self, e, user_cwd):
		self._e = e
		self.user_cwd = user_cwd
		self.project_directory = user_cwd

	def __getattr__(self, name):
		raise self._e

	def __getitem__(self, name):
		return getattr(self, name)

def setup(config_fn=None, debug_cmd=False):
	global cfg
	try:
		locale.setlocale(locale.LC_ALL, '')
	except locale.Error:
		print("WARNING: Broken locale", file=sys.stderr)
	# Make sure the accelerator dir in not in sys.path
	# (as it might be if running without installing.)
	unpath(dirname(__file__))
	if config_fn is False:
		return
	user_cwd = os.getcwd()
	if config_fn:
		load_cfg(config_fn)
	else:
		try:
			load_some_cfg(all=debug_cmd)
		except UserError as e:
			cfg = NoConfig(e, user_cwd)
	cfg.user_cwd = user_cwd
	if not debug_cmd:
		# We want the project directory to be first in sys.path.
		unpath(cfg['project_directory'])
		sys.path.insert(0, cfg['project_directory'])
		# For consistency we also always want the project dir
		# as working directory.
		os.chdir(cfg['project_directory'])


class Command(namedtuple('Command', 'name help is_debug modname')):
	def __call__(self, argv):
		from importlib import import_module
		mod = import_module(self.modname)
		return mod.main(argv, cfg)

_COMMANDS = {}

def add_command(name, help, is_debug=False, modname=None):
	if not modname:
		modname = 'accelerator.shell.' + name.replace('-', '_')
	assert name not in _COMMANDS, name
	_COMMANDS[name] = Command(name, help, is_debug, modname)

add_command('board-server', '''runs a webserver for displaying results''',      modname='accelerator.board')
add_command('ds',           '''display information about datasets''',           is_debug=True)
add_command('gc',           '''delete stale and unused jobs''')
add_command('grep',         '''search for a pattern in one or more datasets''', is_debug=True)
add_command('hist',         '''show a histogram of column(s) from a dataset''', is_debug=True)
add_command('init',         '''create a project directory''')
add_command('job',          '''information about a job''',                      is_debug=True)
add_command('method',       '''information about methods''')
add_command('run',          '''run a build script''',                           modname='accelerator.build')
add_command('script',       '''information about build scripts''')
add_command('sherlock',     '''show which job produced a file''')
add_command('status',       '''server status (like ^T when building)''')
add_command('urd',          '''inspect urd contents''')
add_command('urd-server',   '''run the urd server''',                           modname='accelerator.urd')
add_command('workdir',      '''information about workdirs''', is_debug=True)


def cmd_abort(argv):
	parser = ArgumentParser(prog=argv.pop(0))
	parser.add_argument('-q', '--quiet', action='store_true', negation='not', help="no output")
	args = parser.parse_intermixed_args(argv)
	from accelerator.build import Automata
	a = Automata(cfg.url)
	res = a.abort()
	if not args.quiet:
		print("Killed %d running job%s." % (res.killed, '' if res.killed == 1 else 's'))
cmd_abort.help = '''abort running job(s)'''

def cmd_alias(argv):
	parser = ArgumentParser(
		prog=argv.pop(0),
		description='''shows all aliases with no arguments, or the expansion of the specified aliases.''',
	)
	parser.add_argument('alias', nargs='*', help='')
	args = parser.parse_intermixed_args(argv)
	if args.alias:
		from accelerator.compat import shell_quote
		for alias in args.alias:
			a, b = expand_aliases([], [alias])
			print(' '.join(map(shell_quote, a + b)))
	else:
		for item in sorted(aliases.items()):
			print('%s = %s' % item)
cmd_alias.help = '''show defined aliases'''

def cmd_server(argv):
	from accelerator.server import main
	from accelerator.methods import MethodLoadException
	try:
		main(argv, cfg)
	except MethodLoadException as e:
		print(e)
cmd_server.help = '''run the main server'''

def cmd_intro(argv):
	parser = ArgumentParser(prog=argv.pop(0))
	parser.parse_intermixed_args(argv)
	from accelerator import __version__ as ax_version
	def cmd(txt, *a):
		print('  ' + colour(txt, 'intro/highlight', *a))
	def msg(txt='', c='intro/info'):
		if txt:
			print(colour(txt, c))
		else:
			print()
	msg('Welcome to exax ' + ax_version, 'intro/header')
	msg()
	msg('Run')
	cmd('ax init --examples /tmp/axtest')
	cmd('cd /tmp/axtest')
	msg('to setup a project including example files.')
	msg()
	msg('To see example build scripts, run')
	cmd('ax script')
	msg()
	msg('After starting the server:')
	cmd('ax server')
	msg('try for example the first tutorial script:')
	cmd('ax run tutorial01')
	msg('(The "build_"-prefix is not required.)')
	msg()
	msg('All example code should be in the "examples" directory.')
	msg('All example build scripts will print where they are located.')
	msg()
	msg('To see available methods, run')
	cmd('ax method')
	msg()
	msg('For a longer intro, see')
	cmd('https://exax.org/documentation/2019/10/30/initialise.html', 'intro/info')
cmd_intro.help = '''show introduction text'''

def cmd_version(argv, as_command=True):
	from accelerator import __version__ as ax_version
	if as_command:
		parser = ArgumentParser(prog=argv.pop(0))
		parser.parse_intermixed_args(argv)
	print(ax_version)
	if as_command:
		py_version = ''
		suffix = sys.version.strip()
		try:
			# sys.implementation does not exist in python 2.
			py_version = sys.implementation.name
			suffix = ' (%s)' % (suffix.split('\n')[0].strip(),)
			impl_version = '.'.join(map(str, sys.implementation.version))
			py_version = '%s %s' % (py_version, impl_version,)
		except Exception:
			pass
		print('Running on ' + py_version + suffix)
cmd_version.help = '''show installed accelerator version'''


def split_args(argv):
	prev = None
	for ix, arg in enumerate(argv):
		if not arg.startswith('-') and prev != '--config':
			return argv[:ix], argv[ix:]
		prev = arg
	return argv, []

_unesc_re = re.compile(r'\\([abefnrtv\\]|x[0-9a-f]{2})', re.IGNORECASE)
_unesc_v = {
	'a': '\a',
	'b': '\b',
	'e': '\x1b',
	'f': '\f',
	'n': '\n',
	'r': '\r',
	't': '\t',
	'v': '\v',
	'\\': '\\',
}
def _unesc(m):
	v = m.group(1)
	if len(v) > 1:
		return chr(int(v[1:], 16))
	else:
		return _unesc_v.get(v.lower(), v)

def parse_user_config():
	from accelerator.compat import open
	from configparser import ConfigParser
	from os import environ
	fns = []
	# This can't be an argument, as aliases specified here need to be expanded
	# during argument parsing. Therefore it's an environment variable instead.
	# It's not really intended for users, just for tests.
	if environ.get('ACCELERATOR_IGNORE_ETC', None) is None:
		fns = ['/etc/accelerator/config']
	cfgdir = environ.get('XDG_CONFIG_HOME')
	if not cfgdir:
		home = environ.get('HOME')
		if home:
			cfgdir = join(home, '.config')
	if cfgdir:
		fns.append(join(cfgdir, 'accelerator', 'config'))
	all_cfg = []
	def read(fn, seen_fns=(), must_exist=False):
		try:
			with open(fn, 'r', encoding='utf-8') as fh:
				contents = fh.read()
		except IOError:
			if must_exist:
				raise
			return
		seen_fns = set(seen_fns)
		seen_fns.add(fn)
		cfg = ConfigParser()
		cfg.read_string(contents, fn)
		include = cfg.get('include', 'path', fallback=None)
		if include:
			for include in shlex.split(include):
				include = join(dirname(fn), expanduser(include))
				if include in seen_fns:
					raise Exception('Config include loop: %r goes back to %r' % (fn, include,))
				read(include, seen_fns, True)
		# Append this file after the included files, so this file takes priority.
		all_cfg.append((fn, contents))
	for fn in fns:
		read(fn)
	cfg = ConfigParser()
	cfg.optionxform = str # case sensitive (don't downcase aliases)
	for fn, contents in all_cfg:
		cfg.read_string(contents, fn)
	return cfg

def printdesc(items, columns, colour_prefix, full=False):
	ddot = ' ...'
	def chopline(description, max_len):
		if len(description) > max_len:
			max_len -= len(ddot)
			parts = description.split()
			description = ''
			for part in parts:
				if len(description) + len(part) + 1 > max_len:
					break
				if description:
					description = '%s %s' % (description, part,)
				else:
					description = part
			description += colour.faint(ddot)
		return description
	items = [(name, description.strip('\n').split('\n')) for name, description in items]
	if not full:
		# make names the same length, within same-ish length groups
		lens = set(len(name) for name, _ in items)
		len2len = {}
		group_size = 14
		spaces = ' ' * group_size
		while lens:
			this = min(lens)
			here = {l for l in lens if l < this + group_size}
			m = max(here)
			len2len.update({l: m for l in here})
			lens -= here
		items = [
			(
				(name + spaces)[:len2len[len(name)]],
				[description[0]],
			)
			for name, description in items
		]
	items = [(name, description if description[0] else None) for name, description in items]
	for name, description in items:
		max_len = columns - len(ddot) - len(name)
		preamble = colour('  ' + name, colour_prefix + '/highlight')
		if description and max_len > 10:
			if full:
				print(preamble)
				for line in description:
					print('    ' + line)
			else:
				print(preamble + '  ' + chopline(description[0], max_len))
		else:
			print(preamble)

def expand_env(words, alias):
	from os import environ
	for word in words:
		if word.startswith('${') and word.endswith('}'):
			k = word[2:-1]
			if k in environ:
				try:
					expanded = shlex.split(environ[k])
				except ValueError as e:
					raise ValueError('Failed to expand alias %s (%s -> %r): %s' % (alias, word, environ[k], e,))
				for word in expanded:
					yield word
		else:
			yield word

def expand_aliases(main_argv, argv):
	used_aliases = []
	while argv and argv[0] in aliases:
		alias = argv[0]
		if alias == 'noalias': # save the user from itself
			break
		try:
			expanded = shlex.split(aliases[alias])
		except ValueError as e:
			raise ValueError('Failed to expand alias %s (%r): %s' % (argv[0], aliases[argv[0]], e,))
		expanded = list(expand_env(expanded, alias))
		more_main_argv, argv = split_args(expanded + argv[1:])
		main_argv.extend(more_main_argv)
		if expanded and alias == expanded[0]:
			break
		used_aliases.append(alias)
		if alias in used_aliases[:-1]:
			raise ValueError('Alias loop: %r' % (used_aliases,))

	while argv and argv[0] == 'noalias':
		argv.pop(0)
	return main_argv, argv

def main():
	# Several commands use SIGUSR1 which (naturally...) defaults to killing the
	# process, so start by blocking that to minimise the race time.
	if hasattr(signal, 'pthread_sigmask'):
		signal.pthread_sigmask(signal.SIG_BLOCK, {signal.SIGUSR1})
	else:
		# Or if we can't block it, just ignore it.
		signal.signal(signal.SIGUSR1, signal.SIG_IGN)

	# As of python 3.8 the default start_method is 'spawn' on macOS.
	# This doesn't work for us. 'fork' is fairly unsafe on macOS,
	# but it's better than not working at all. See
	# https://bugs.python.org/issue33725
	# for more information.
	import multiprocessing
	if hasattr(multiprocessing, 'set_start_method'):
		# If possible, make the forkserver (used by database updates) pre-import everthing
		if hasattr(multiprocessing, 'set_forkserver_preload'):
			multiprocessing.set_forkserver_preload(['accelerator', 'accelerator.server'])
		multiprocessing.set_start_method('fork')

	from accelerator import g
	g.running = 'shell'

	from accelerator.autoflush import AutoFlush
	main_argv, argv = split_args(sys.argv[1:])
	sys.stdout = AutoFlush(sys.stdout)
	sys.stderr = AutoFlush(sys.stderr)

	# configuration defaults
	global aliases
	aliases = {
		'cat': 'grep -e ""',
	}
	colour_d = {
		'warning': ('RED',),
		'highlight': ('BOLD',),
		'grep/highlight': ('RED',),
		'info': ('BRIGHTBLUE',),
		'infohighlight': ('BOLD', 'BRIGHTBLUE',),
		'separator': ('CYAN', 'UNDERLINE',),
		'lined/separator': (),
		'bar': ('BRIGHTBLUE',),
		'count': (),
		'range': ('BOLD',),
		'empty': ('RESET', 'FAINT',), # many terminals ignore faint, so reset to un-bold.
		'header': ('BRIGHTBLUE', 'BOLD',),
		'evenlines': ('BLACK', 'WHITEBG',),
		'oddlines': ('BLACK', 'BRIGHTWHITEBG',),
	}

	global user_cfg
	user_cfg = parse_user_config()
	if 'alias' in user_cfg:
		aliases.update(user_cfg['alias'])
	if 'colour' in user_cfg:
		colour_d.update({k: [_unesc_re.sub(_unesc, e) for e in v.split()] for k, v in user_cfg['colour'].items()})
	colour._names.update(colour_d)

	main_argv, argv = expand_aliases(main_argv, argv)

	# any function cmd_* in this file is a command
	COMMANDS = {
		k[4:].replace('_', '-'): v
		for k, v in globals().items()
		if k.startswith('cmd_') and callable(v)
	}
	# as is anything in _COMMANDS (i.e. from add_command)
	assert not set(COMMANDS).intersection(_COMMANDS)
	COMMANDS.update(_COMMANDS)

	epilog = ['commands:', '']
	cmdlen = max(len(cmd) for cmd in COMMANDS)
	template = '  %%%ds  %%s' % (cmdlen,)
	for cmd, func in sorted(COMMANDS.items()):
		epilog.append(template % (cmd, func.help,))
	epilog.append('')
	epilog.append('aliases:')
	epilog.extend('  %s = %s' % item for item in sorted(aliases.items()))
	epilog.append('')
	epilog.append('use "' + colour('%(prog)s <command> --help', 'help/highlight') + '" for <command> usage')
	epilog.append('try "' + colour('%(prog)s intro', 'help/highlight') + '" for an introduction')
	parser = ArgumentParser(
		usage='%(prog)s [--config CONFIG_FILE] command [args]',
		epilog='\n'.join(epilog),
		formatter_class=RawDescriptionHelpFormatter,
	)
	parser.add_argument('--config', metavar='CONFIG_FILE', help='configuration file')
	parser.add_argument('--version', action='store_true', negation='no', help='alias for the version command')
	args = parser.parse_intermixed_args(main_argv)
	if args.version:
		sys.exit(cmd_version((), False))
	args.command = argv.pop(0) if argv else None
	if args.command not in COMMANDS:
		parser.print_help(file=sys.stderr)
		if args.command is not None:
			print(file=sys.stderr)
			print('Unknown command "%s"' % (args.command,), file=sys.stderr)
		sys.exit(2)
	config_fn = args.config
	if args.command in ('init', 'intro', 'version', 'alias',):
		config_fn = False
	cmd = COMMANDS[args.command]
	debug_cmd = getattr(cmd, 'is_debug', False)
	try:
		setup(config_fn, debug_cmd)
		argv.insert(0, '%s %s' % (basename(sys.argv[0]), args.command,))
		return cmd(argv)
	except UserError as e:
		print(e, file=sys.stderr)
		return 1
	except OSError as e:
		if e.errno == errno.EPIPE:
			return 1
		else:
			raise
	except KeyboardInterrupt:
		# Exiting with KeyboardInterrupt causes python to print a traceback.
		# We don't want that, but we do want to exit from SIGINT (so the
		# calling process can know that happened).
		signal.signal(signal.SIGINT, signal.SIG_DFL)
		os.kill(os.getpid(), signal.SIGINT)
		# If that didn't work let's re-raise the KeyboardInterrupt.
		raise



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Test the dataset_concat method.
'''

from datetime import date, time, datetime
from itertools import chain

from accelerator import subjobs, JobError
from accelerator.compat import PY2
from accelerator.dsutil import _type2iter

def synthesis(job):
	# Test using all types
	types = {
		'a': 'ascii',
		'b': 'bool',
		'c': 'bytes',
		'd': 'complex32',
		'e': 'complex64',
		'f': 'date',
		'g': 'datetime',
		'h': 'float32',
		'i': 'float64',
		'j': 'int32',
		'k': 'int64',
		'l': 'json',
		'm': 'number',
		'n': 'pickle',
		'o': 'time',
		'p': 'unicode',
	}
	missing = set(_type2iter) - set(types.values())
	assert not missing, missing
	if PY2:
		del types['n'] # no pickle type on python2

	def data(ix):
		d = {
			'a': '%d' % (ix,),
			'b': bool(ix % 2),
			'c': b'%d' % (ix,),
			'd': complex(0, ix),
			'e': complex(0, -ix),
			'f': date(2000 + ix // 12, ix % 12 + 1, 1),
			'g': datetime.fromtimestamp(ix),
			'h': 1.0 / (ix + 1),
			'i': 2.0 / (ix + 1),
			'j': ix,
			'k': -ix,
			'l': {'json': ix},
			'm': -1.0 / (ix + 1),
			'n': {ix},
			'o': time(0, ix // 60 % 60, ix % 60),
			'p': u'%d' % (ix,),
			'extra': 0,
		}
		return {k: v for k, v in d.items() if k in types}

	# default to hashlabel='b' so that only two slices get data
	def mk_ds(name, num, hashlabel='b', empty=False, **kw):
		dw = job.datasetwriter(name=name, columns=types, hashlabel=hashlabel, **kw)
		write = dw.get_split_write()
		if not empty:
			for ix in range(100 * num, 100 * num + 100):
				write(**data(ix))
		return dw.finish()

	previous = None
	for num in range(8):
		previous = mk_ds(str(num), num, previous=previous)
	# we want one empty to make sure nothing breaks from that
	empty = mk_ds('empty', 0, empty=True, previous=previous)
	assert len(list(empty.iterate(None))) == 0
	last = mk_ds('8', 8, previous=empty)
	numbered = last.chain()
	assert len(numbered) == 10

	# Verify contents, chaining and types (including none_support)
	def chk(source, previous, want_in_chain, want, do_sort=True, none_support=()):
		ds = subjobs.build('dataset_concat', source=source, previous=previous).dataset()
		assert ds.chain() == want_in_chain + [ds]
		want = list(chain(*[w.iterate(None) for w in want]))
		got = list(ds.iterate(None))
		if do_sort:
			got.sort()
			want.sort()
		assert want == got, source
		want_types = {k: (v, k in none_support) for k, v in types.items()}
		got_types = {k: (v.type, v.none_support) for k, v in ds.columns.items()}
		assert want_types == got_types, source
		return ds

	a = chk(numbered[3], None, [], numbered[:4])
	b = chk(numbered[6], a, [a], numbered[4:7])
	# c uses just one dataset, so order should not change from source
	c = chk(numbered[7], b, [a, b], [numbered[7]], do_sort=False)
	# make sure an empty dataset alone is fine
	chk(empty, c, [a, b, c], [empty])
	# and that an empty dataset as part of the chain is fine
	chk(last, b, [a, b], numbered[7:])
	# empty last in chain, equivalent to c above
	chk(empty, b, [a, b], [numbered[7]], do_sort=False)
	# empty first in chain
	chk(last, c, [a, b, c], numbered[8:], do_sort=False)

	# Test some things that should fail
	def want_fail(why, **kw):
		try:
			subjobs.build('dataset_concat', **kw)
			raise Exception('dataset_concat(%r) should have failed: %s' % (kw, why,))
		except JobError:
			pass

	bad_hl = mk_ds('bad_hl', 8, hashlabel='a', previous=numbered[7])
	want_fail('differing hashlabels', source=bad_hl)

	types['extra'] = 'int64'
	extra_col = mk_ds('extra_col', 8, previous=numbered[7])
	want_fail('extra column', source=extra_col)
	del types['extra']

	del types['k']
	missing_col = mk_ds('missing_col', 8, previous=numbered[7])
	want_fail('column k missing', source=missing_col)

	types['k'] = 'number' # should be int64
	bad_type = mk_ds('bad_type', 8, previous=numbered[7])
	want_fail('different types on column k', source=bad_type)

	# Test that having a none_support somewhere is inherited
	types['k'] = ('int64', True)
	with_none = mk_ds('with_none', 8, previous=numbered[7])
	types['k'] = 'int64'
	without_none = mk_ds('without_none', 9, previous=with_none)
	chk(without_none, c, [a, b, c], [with_none, without_none], none_support="k")
